#!/usr/bin/perl 
# Store the disk ipl program on a premarked X1210/1215 disk pack image 
# (in its predefined location is logical sector 1 / physical sector 3)

# usage:
# ipl2disk <disk image file name>.img <ipl file name>.abs

use strict;
use bytes;
my $DEBUG = 1;

###############################################################
# disk pack parameters
my $type  = 'XM'; # disk unit type
my $cyls  = 203;  # number of tracks = number of cylinders
my $heads = 2;    # number of heads
my $spt   = 16;   # number of sectors per track
my $lsec  = 205;  # sector length is 205 16 bit words
#
my $cyl;          # working variable
my $head;         # working variable
my $sector;       # working variable
###############################################################
# Each sector consists of 205 16 bit data words, 5 overhead and 200 for user data.
#
# Sector format:
use constant IDENT => 0x0; # word 1 is the sector identifier
use constant SEQM  => 0x2; # word 2 is for sequential file admin
use constant REC   => 0x4; # word 3 is start of the data part
#   xxxxxxxx 200 data words
#   xxxxxxxx 3   reserved words
#
# physical sector  logical sector
#    sector 0   <->   sector 0
#    sector 1   <->   sector 11
#    sector 2   <->   sector 6
#    sector 3   <->   sector 1
#    sector 4   <->   sector 12
#    sector 5   <->   sector 7
#    sector 6   <->   sector 2
#    sector 7   <->   sector 13
#    sector 8   <->   sector 8
#    sector 9   <->   sector 3
#    sector 10  <->   sector 14
#    sector 11  <->   sector 9
#    sector 12  <->   sector 4
#    sector 13  <->   sector 15
#    sector 14  <->   sector 10
#    sector 15  <->   sector 5
#
# physical-sector-in-track = remainder( modulo-16 (logical-sector-in-track * 3))
#
my $n = $cyls*$heads*$spt;                # number of sectors on disk
my $g = int($n/8);                        # number of granules

# An X1210 disk contains 203 * 2 * 16 = 6496 sectors
# An X1210 disk contains 6496/8       = 812  granules
#
#
# Disk layout
# granule 0
#   sector0 = volume label/granule allocation table
#   sector1 = IPL
#   sector2..7 = catalogue
#      each catalogue entry takes 8 16-bit words
#      maximum = 6 * 25 = 150 catelogue entries (each entry pointing to a directory)
#      The 1st catalogue entry is for the system and it points to the system 
#      directory with system software.
#      The catalogue ends with an all one 16 bit word (0xFFFF)
# granule 1 = system directory
#      each directory entry takes 8 16-bit words
#      maximum = 8 * 25 = 200 directory entries (each entry pointing to a file)
#      The system directory starts at (logical) sector 8
#      The 1st directory entry is the monitor (sector 16)
#      The directory ends with an all one 16 bit word (0xFFFF)
#      User directories use the same format
# granule 2..811 = monitor, user directories and files
#      the monitor file takes consecutive granules 2..n (on a system disk)
#      the monitor load module starts at sector 16
# NOTE:on the P800 system, the monitor was stored on the diskpack during
#      system generation, which also allocated/reserved space on the disk
#      pack for that monitor. In this PREMARK we use a different approach:
#      just 8 granules (granule 2..9, good for a load mudule file of ~ 11.5 
#      kwords) are reseved for storing a monitor load module file, starting at 
#      granule 2 (logical sector 16). This monitor is meant to be loaded and 
#      started by the IPL in logical sector 1 of the disk.
#

my $pos;                                  # seek position
my $i=0;
my $j=1;
my $w;
my $nbrw=0;                               # number of words written on disk

print "P860 X1210/X1215 IPL file loader on a premarked disk pack\n";

my $disk = shift (@ARGV);
my $file = shift (@ARGV);

if ($disk eq "" || $disk !~ /\.img/ || $file eq "" || $file !~ /\.abs/)
{
    print "Parameter error\n";
	print "Usage: mon2disk.pl <diskimagename>.img <iplfilename>.abs\n";
    exit;
}

open (FDISK, "+<:raw", $disk) || die "cannot open image file $disk: $!";
print "X1210/X1215 moving head disk pack: $disk\n";
print "Load IPL file: $file on disk pack $disk\n";

open (IN, $file) || die "cannot open load module file $file: $!";
binmode IN, ":raw";

while(rc() != 0xff)
{
	next;
}

# Read 2 words of 8*8 papertape (word count and startaddr-2)
rw(); # 0
rw(); # 0

my $ls = 1;                     # logical sector to load
if(lseek($ls))	                   # seek to data part of requested sector
{
	while($i<200) # copy code/data words from papertape to disk
	{
		$w=rw();
		ww($w);
		$i=$i+1;
	}
}
else
{
	print "Seek error when starting to write logical sector 0x12\n";
}
rw(); #checsum

# last byte of papertape
if(rc() != 0x94)
{
	print "Error while reading IPL form papertape\n";
}

close FDISK || die "can't close $disk: $!";	
close IN    || die "can't close $file: $!";
print "IPL load on disk ready\n";
print "$nbrw words written to disk\n";

####################################################################

# seek to the DATA part of a logical sector
sub lseek {
	my $s = shift;                 # logical disk sector number
    my $c = int($s/($spt*$heads)); # cylinder
	my $r = $s%($spt*$heads);
	my $h = int($r/$spt);          # head
	my $l = $r%$spt;               # logical sector in track
	my $p = secphys($l);           # physical sector in track
	
	$pos = $c*$spt*$heads*$lsec*2 + $h*$spt*$lsec*2 + $p*$lsec*2 + REC;		
	if(sysseek(FDISK, $pos, 0) eq undef)
	{
		print "Seek error\n";
		return undef
	}
	return $pos;
}

# read next byte from papertape
sub rc {
		my $c;
	    read(IN,$c,1);
		$c = unpack("C", $c); 	            # get new current char
		return ($c & 0xFF);
}

# read next word from papertape
sub rw {
	my $c;
	$w = rc();
	$w = $w << 8;
	$c = rc();
	$w = $w + $c;
	return $w;
}

# write next word to disk
sub ww {
	my $w = shift;
	wb(($w >> 8) & 0xFF);
	wb($w & 0xFF);
	$nbrw=$nbrw+1;                           # number of words written
}

# write next byte to disk
sub wb {
	my $b =shift;
	$b = pack("C", $b);
	if(syswrite(FDISK, $b, 1) != 1)
	{
		print "Disk write error\n";
	}
}

# convert a logical sector number in track to a physical sector number in track
sub secphys {
	my $sl = shift;
	$sl=$sl*3;
	if($sl > 16)
	{
		$sl=$sl-16;
		if($sl > 16)
		{
			$sl=$sl-16;
		}
	}
	return $sl;
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf("%01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf("%02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf("%03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf("%04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf("%05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf("%06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf("%07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf("%08X", $v);
	}
	# print "$s ";
	return $s;
}
